<?php
namespace app\modules\organization\application;

use app\common\base\Model;
use app\common\data\ActiveDataProvider;
use app\common\db\ActiveRecord;
use app\common\helpers\ArrayHelper;
use app\common\helpers\CommonHelper;
use app\common\helpers\Json;
use app\common\logic\orm\Address;
use app\common\logic\orm\Email;
use app\common\logic\orm\Phone;
use app\common\service\ApplicationService;
use app\common\service\exception\AccessApplicationServiceException;
use app\common\service\exception\ApplicationServiceException;
use app\modules\organization\models\finders\EmployeeFinder;
use app\modules\organization\models\orm\Employee;
use yii\base\InvalidValueException;
use app\modules\organization\models\form\Employee as EmployeeForm;

/**
 * Class EmployeeService
 * @package app\modules\organization\application
 *
 * 
 */
class EmployeeService extends ApplicationService implements EmployeeServiceInterface
{
    /**
     * @inheritdoc
     */
    public function aclAlias()
    {
        return 'Сотрудник';
    }

    /**
     * @inheritdoc
     */
    public function getPrivileges()
    {
        return [
            'addEmployee' => 'Добавить сотрудника',
            'updateEmployee' => 'Обновить сотрудника',
            'getEmployeeById' => 'Просмотр сотрудника',
            'getEmployeeList' => 'Список сотрудников',
        ];
    }

    /**
     * @inheritdoc
     */
    public function init()
    {
        $this->modelClass = Employee::class;
        parent::init();
    }

    /**
     * @inheritdoc
     */
    public function getEmployeeList(Model $form)
    {
        /** @var $form EmployeeFinder */
        if (!$this->isAllowed('getEmployeeList')) {
            throw new AccessApplicationServiceException('Доступ к списку сотрудников запрещён.');
        }
        $query = Employee::find();
        $query
            ->andFilterWhere([
                'cast(updated_at as date)' =>
                    empty($form->updatedAt) ? null : \Yii::$app->formatter->asDate($form->updatedAt, CommonHelper::FORMAT_DATE_DB),
            ]);
        return new ActiveDataProvider([
            'query' => $query,
            'pagination' => [
                'pageSize' => 10
            ],
        ]);
    }

    /**
     * @param string $date
     * @return array
     */
    public function getEmployeesWithAttendanceByDate(string $date)
    {
        $employees = Employee::find()
            ->joinWith([
                'attendances.ehr.patient',
                'speciality'
            ])
//            ->where([
//                'cast(datetime as DATE)' => $date, // todo CAST субд
//            ])
            ->all();
        return $employees;
    }

    /**
     * @inheritdoc
     */
    public function getEmployeeById($id)
    {
        return Employee::findOneEx($id);
    }

    public function addEmployee(EmployeeForm $employeeForm)
    {
        if (!($employeeForm instanceof EmployeeForm)) {
            throw new InvalidValueException('Form is not instance ' . EmployeeForm::class . ' class');
        }
        $transaction = \Yii::$app->db->beginTransaction();
        try {
            $model = new Employee([
                'scenario' => ActiveRecord::SCENARIO_CREATE
            ]);
            $model->loadForm($employeeForm);
            if (!$model->save()) {
                throw new ApplicationServiceException('Не удалось сохранить сотрудника.');
            }
            $employeeForm->id = $model->id;
//            $this->savePositions($model->id, $employeeDto->positions);
            $this->savePhones($model->id, $employeeForm->phones);
            $this->saveEmails($model->id, $employeeForm->emails);
            $this->saveAddresses($model->id, $employeeForm->addresses);
            $transaction->commit();
        } catch (\Exception $e) {
            $transaction->rollBack();
            throw $e;
        }
        return $model;
    }

    public function updateEmployee($id, EmployeeForm $employeeForm)
    {
        if (!($employeeForm instanceof EmployeeForm)) {
            throw new InvalidValueException('Form is not instance ' . EmployeeForm::class . ' class');
        }
        $transaction = \Yii::$app->db->beginTransaction();
        try {
            /** @var ActiveRecord $model */
            $model = Employee::findOneEx($id);
            $model->setScenario(ActiveRecord::SCENARIO_UPDATE);
            $model->loadForm($employeeForm);
            if (!$model->save()) {
                throw new ApplicationServiceException('Не удалось сохранить сотрудника.');
            }
            $employeeForm->id = $model->id;
//            $this->savePositions($model->id, $employeeDto->positions);
            $this->savePhones($model->id, $employeeForm->phones);
            $this->saveEmails($model->id, $employeeForm->emails);
            $this->saveAddresses($model->id, $employeeForm->addresses);
            $transaction->commit();
        } catch (\Exception $e) {
            $transaction->rollBack();
            throw $e;
        }
        return $model;
    }

    /**
     * @param string $employeeId
     * @param array $emails
     * @return void
     * @throws ApplicationServiceException
     */
    protected function saveEmails($employeeId, $emails)
    {
        $employee = Employee::findOneEx($employeeId);
        $q = Email::find()
            ->where([
                'entity' => Employee::getTableSchema()->name,
                'entity_id' => $employee->id
            ])
            ->notDeleted();
        $exists = $q->all();
        foreach ($exists as $e) {
            $e->delete();
        }
        if (!is_array($emails)) {
            return null;
        }
        foreach ($emails as $email) {
            if (empty($email['type']) || empty($email['address'])) {
                continue;
            }
            $model = new Email();
            $model->setAttributes($email);
            $model->entity = Employee::getTableSchema()->name;
            $model->entity_id = $employee->id;
            if (!$model->save()) {
                throw new ApplicationServiceException('Не удалось сохранить email. Причина: ' . Json::encode($model->getErrors()));
            }
        }
    }

    /**
     * @param string $employeeId
     * @param array $phones
     * @return void
     * @throws ApplicationServiceException
     */
    protected function savePhones($employeeId, $phones)
    {
        $employee = Employee::findOneEx($employeeId);
        $q = Phone::find()
            ->where([
                'entity' => Employee::getTableSchema()->name,
                'entity_id' => $employee->id
            ])
            ->notDeleted();
        $exists = $q->all();
        foreach ($exists as $p) {
            $p->delete();
        }
        if (!is_array($phones)) {
            return null;
        }
        foreach ($phones as $phone) {
            if (empty($phone['type']) || empty($phone['phone'])) {
                continue;
            }
            $model = new Phone();
            $model->setAttributes($phone);
            $model->entity = Employee::getTableSchema()->name;
            $model->entity_id = $employee->id;
            if (!$model->save()) {
                throw new ApplicationServiceException('Не удалось сохранить телефон. Причина: ' . Json::encode($model->getErrors()));
            }
        }
    }

    /**
     * @param string $employeeId
     * @param array $addresses
     * @return void|null
     */
    protected function saveAddresses($employeeId, $addresses)
    {
        $employee = Employee::findOneEx($employeeId);
        $q = Address::find()
            ->where([
                'entity' => Employee::getTableSchema()->name,
                'entity_id' => $employee->id
            ])
            ->notDeleted();
        $exists = $q->all();
        foreach ($exists as $p) {
            $p->delete();
        }
        if (!is_array($addresses)) {
            return null;
        }
        foreach ($addresses as $address) {
            if (empty($address['type']) || empty($address['city'])) {
                continue;
            }
            $model = new Address();
            $model->setAttributes($address);
            $model->entity = Employee::getTableSchema()->name;
            $model->entity_id = $employee->id;
            $model->save();
        }
    }

//    public function savePositions($employeeId, $positionIds)
//    {
//        $employee = Employee::findOneEx($employeeId);
//        $q = PositionEmployee::find()
//            ->where([
//                'employee_id' => $employee->id
//            ])
//            ->notDeleted();
//        $positionEmployees = $q->all();
//        if (!empty($positionEmployees)) {
//            foreach ($positionEmployees as $p) {
//                $p->delete();
//            }
//        }
//        if (empty($positionIds) && !ArrayHelper::isAssociative($positionIds)) {
//            return null;
//        }
//        foreach ($positionIds as $positionId) {
//            /** @var ActiveQuery $q */
//            $q = Position::find()
//                ->where([
//                    'id' => $positionId,
//                ])
//                ->notDeleted();
//            /** @var Position $position */
//            $position = $q->one();
//            if ($position === null) {
//                throw new ServiceException('Position is not found in db.');
//            }
//            $relation = new PositionEmployee();// @todo лучше ->load
//            $relation->position_id = $position->id;
//            $relation->employee_id = $employee->id;
//            $relation->save();
//        }
//    }

    /**
     * @param string $raw
     * @return EmployeeForm
     */
    public function getEmployeeForm($raw)
    {
        $model = Employee::ensureWeak($raw);
        $employeeForm = new EmployeeForm();
        if ($model->isNewRecord) {
            $employeeForm->setScenario('create');
        }
        $employeeForm->loadAr($model);
        $employeeForm->id = $model->id;
        $employeeForm->phones = ArrayHelper::toArray($model->phones);
        $employeeForm->emails = ArrayHelper::toArray($model->emails);
        $employeeForm->addresses = ArrayHelper::toArray($model->addresses);
        $employeeForm->speciality = ArrayHelper::toArray(!$model->speciality ? [] : $model->speciality);
        $employeeForm->user = ArrayHelper::toArray(!$model->user ? [] : $model->user);
        return $employeeForm;
    }
}
