<?php
namespace app\modules\security\application;

use app\common\acl\resource\ApplicationResourceInterface;
use app\common\base\Model;
use app\common\base\Module;
use app\common\data\ActiveDataProvider;
use app\common\db\ActiveRecord;
use app\common\dto\Dto;
use app\common\helpers\ClassHelper;
use app\common\helpers\CommonHelper;
use app\common\helpers\Json;
use app\common\service\ApplicationService;
use app\common\service\ApplicationServiceInterface;
use app\common\service\exception\AccessApplicationServiceException;
use app\common\service\exception\ApplicationServiceException;
use app\common\service\ServiceRegistry;
use app\common\workflow\WorkflowRegistry;
use app\modules\security\models\finders\AclFinder;
use app\modules\security\models\orm\Acl;
use app\modules\security\models\orm\AclRole;
use yii\base\InvalidValueException;

/**
 * Class AclAccessService
 * @package app\modules\security\application
 *
 *
 */
class AclService extends ApplicationService implements AclServiceInterface, ApplicationResourceInterface
{
    /**
     * @var string
     */
    public $modelClass = Acl::class;

    /**
     * @inheritdoc
     */
    public function aclAlias()
    {
        return 'ACL';
    }

    /**
     * @inheritdoc
     */
    public function getPrivileges()
    {
        return [
            'add' => 'Создание ACL',
            'update' => 'Обновление ACL',
            'getAclList' => 'Список ACL',
            'getAclRoleList' => 'Список ролей',
        ];
    }

    /**
     * @inheritdoc
     */
    public function getAclList(Model $form)
    {
        /** @var $form AclFinder */
        if (!$this->isAllowed('getAclList')) {
            throw new AccessApplicationServiceException('Доступ к списку ACL запрещён.');
        }
        $query = Acl::find();
        $query
            ->andFilterWhere([
                'cast(updated_at as date)' =>
                    empty($form->updatedAt) ? null : \Yii::$app->formatter->asDate($form->updatedAt, CommonHelper::FORMAT_DATE_DB),
            ])
            ->orderBy(['module' => SORT_ASC, 'entity_type' => SORT_ASC]);
        return new ActiveDataProvider([
            'query' => $query,
            'pagination' => [
                'pageSize' => 10
            ],
        ]);
    }

    /**
     * @inheritdoc
     */
    public function getAclRoleList(Model $form)
    {
        /** @var $form AclFinder */
        if (!$this->isAllowed('getAclRoleList')) {
            throw new AccessApplicationServiceException('Доступ к списку ролей запрещён.');
        }
        $query = AclRole::find();
        $query
            ->andFilterWhere([
                'cast(updated_at as date)' =>
                    empty($form->updatedAt) ? null : \Yii::$app->formatter->asDate($form->updatedAt, CommonHelper::FORMAT_DATE_DB),
            ]);
        return new ActiveDataProvider([
            'query' => $query,
            'pagination' => [
                'pageSize' => 10
            ],
        ]);
    }

    /**
     * @param Dto $aclDto
     * @param string $scenario
     * @return ActiveRecord
     * @throws \Exception
     */
    public function add($aclDto, $scenario = ActiveRecord::SCENARIO_CREATE)
    {
        if (!$this->isAllowed('add')) {
            throw new AccessApplicationServiceException('Доступ к созданию ACL запрещён.');
        }
        $modelClass = $this->modelClass;
        if (!($aclDto instanceof Dto)) {
            throw new InvalidValueException('object is not instance Dto class'); // todo normalize text
        }
        $transaction = \Yii::$app->db->beginTransaction();
        try {
            /** @var ActiveRecord $model */
            $model = new $modelClass([
                'scenario' => $scenario
            ]);
            $model->loadDto($aclDto);
            if (!$model->save()) {
                $errors = Json::encode($model->getErrors());
                throw new ApplicationServiceException('Не удалось сохранить ACL. Причина: ' . $errors);
            }
            $transaction->commit();
        } catch (\Exception $e) {
            $transaction->rollBack();
            throw $e;
        }
        return $model;
    }

    /**
     * @param string $id
     * @param Dto $aclDto
     * @param string $scenario
     * @return ActiveRecord
     * @throws \Exception
     */
    public function update($id, $aclDto, $scenario = ActiveRecord::SCENARIO_UPDATE)
    {
        if (!$this->isAllowed('update')) {
            throw new AccessApplicationServiceException('Доступ к обновлению ACL запрещён.');
        }
        $modelClass = $this->modelClass;
        if (!($aclDto instanceof Dto)) {
            throw new InvalidValueException('object is not instance Dto class'); // todo normalize text
        }
        $transaction = \Yii::$app->db->beginTransaction();
        try {
            /** @var ActiveRecord $modelClass */
            $model = $modelClass::findOneEx($id);
            $model->setScenario($scenario);
            $model->loadDto($aclDto);
            if (!$model->save()) {
                $errors = Json::encode($model->getErrors());
                throw new ApplicationServiceException('Не удалось сохранить ACL. Причина: ' . $errors);
            }
            $transaction->commit();
        } catch (\Exception $e) {
            $transaction->rollBack();
            throw $e;
        }
        return $model;
    }

    public function deleteAcl($id)
    {
        $acl = Acl::findOneEx($id);
        return $acl->delete();
    }

    /**
     * @param string $module
     * @param int $aclType
     * @return array
     * @throws \Exception
     */
    public static function resourceRegistry($module, $aclType)
    {
        $services = [];
        if ($aclType === Acl::TYPE_SERVICE) {
            $services = ServiceRegistry::registry($module, false);
        } elseif ($aclType === Acl::TYPE_WORKFLOW) {
            $services = WorkflowRegistry::registry($module, false);
        }
        $aliases = [];
        foreach ($services as $entityType) {
            $service = '';
            if ($aclType === Acl::TYPE_SERVICE) {
                $service = ServiceRegistry::getNamespace($module, $entityType);
            } elseif ($aclType === Acl::TYPE_WORKFLOW) {
                $service = WorkflowRegistry::getNamespace($module, $entityType);
            }
            if (!class_exists($service)) {
                throw new \Exception('Класс не найден.');
            }
            if (!ClassHelper::implementsInterface($service, ApplicationServiceInterface::class)) { // @todo ресурс не только App
                continue;
            }
            $aliases[$entityType] =  \Yii::createObject($service)->aclAlias();
        }
        return $aliases;
    }

    public static function getResourceClass($module, $className, $aclType)
    {
        if (class_exists($className)) {
            return $className;
        }
        /** @var Module $m */
        $m = \Yii::$app->getModule($module);
        if ($m === null) {
            throw new ApplicationServiceException('Модуль не найден.');
        }
        $parentModule = $m->getUniqueId();
        $subModules = $m->getModules();
        if (!empty($subModules)) {
            // recursive
            foreach ($subModules as $subModule => $val) {
                if ($aclType === Acl::TYPE_SERVICE) {
                    $modelClass = ServiceRegistry::getNamespace($parentModule . '/' . $subModule, $className);
                } elseif ($aclType === Acl::TYPE_WORKFLOW) {
                    $modelClass = WorkflowRegistry::getNamespace($parentModule . '/' . $subModule, $className);
                }
                if (!class_exists($modelClass)) {
                    continue;
                }
                return $modelClass;
            }
        }
        if ($aclType === Acl::TYPE_SERVICE) {
            $serviceNamespace = $m->getApplicationNamespace();
        } elseif ($aclType === Acl::TYPE_WORKFLOW) {
            $serviceNamespace = $m->getWorkflowNamespace();
        }
        $modelClass = $serviceNamespace . '\\' . $className;
        if (!class_exists($modelClass)) {
            return null;
        }
        return $modelClass;
    }
}
